package com.farata.messaging.channels
{
	import com.farata.messaging.messages.ReliableClientMessage;
	import com.farata.messaging.messages.ReliableServerMessage;
	
	import flash.events.IOErrorEvent;
	import flash.utils.Dictionary;
	import flash.utils.setInterval;
	import flash.utils.setTimeout;
	
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.messaging.MessageAgent;
	import mx.messaging.MessageResponder;
	import mx.messaging.channels.RTMPChannel;
	import mx.messaging.events.MessageAckEvent;
	import mx.messaging.events.MessageEvent;
	import mx.messaging.messages.IMessage;

	public class QoSRTMPChannel extends mx.messaging.channels.RTMPChannel {
		public static const RESEND_TIMEOUT:int = 3000;   // Resend unconfirmed message after 3 sec
		public static const RESENDER_SLEEP_INTERVAL:int = 500; // 

		public function QoSRTMPChannel(id:String=null, uri:String=null)	{
			super(id, uri);
			incoming = new Dictionary();
			unconfirmed = new Dictionary();
			setInterval( resendNonDelivered , RESENDER_SLEEP_INTERVAL);
		}
		override protected function ioErrorHandler( event :IOErrorEvent ) : void {
			super.ioErrorHandler( event );
			setTimeout(resendNonDelivered, 1);
		}

		public function resendNonDelivered() : void {
			for each (var envelope:Object in unconfirmed) {
				if ( new Date().valueOf() - envelope.registeredTs > RESEND_TIMEOUT) {
					var message:IMessage = envelope.message;
					message.headers["duplicate"]="true";  // non-functional, for tracing only
					send( envelope.agent, message);
				}
			}
		}

		override public function send( agent : MessageAgent, message:IMessage) : void {
			if (message is ReliableClientMessage) {
				 unconfirmed[ message.messageId] = {
				 	message:message,
					registeredTs: new Date().valueOf(),
					agent:agent
				 };
				// Emulate 20% of "lost" messages
				if (( message.headers["tm"] != null) && (Math.random()<.2)) {
					if (Log.isDebug()) logger.debug(
						"ResendingChannel emulates loss of message " + message.headers["seqNo"]
					);
					return;
				} else { 
					if (Log.isDebug()) logger.debug(
						"ResendingChannel sends through message " + message.headers["seqNo"]
					);
					super.send( agent, message );
				}
			} else 
				super.send( agent, message );
		}

		private var _defaultAgentListener:Dictionary = new Dictionary();
		override   protected function getDefaultMessageResponder(agent:MessageAgent, msg:IMessage):MessageResponder
    	{
    		if (msg is ReliableClientMessage) {
    			if (agent != null && _defaultAgentListener[agent] == null ) {
    				_defaultAgentListener[agent] = agent;
    				agent.addEventListener(MessageAckEvent.ACKNOWLEDGE,  onProducerAcknowledge );
    			}
    		}
        	return super.getDefaultMessageResponder(agent, msg);
    	}
		private function onProducerAcknowledge(event:MessageEvent):void{
			var ackEvent:MessageAckEvent = event as MessageAckEvent;
 			if (ackEvent && unconfirmed[ackEvent.correlationId]) {
 				var message:ReliableClientMessage = unconfirmed[ackEvent.correlationId].message; 
				if (Log.isDebug()) logger.debug(
					"ResendingChannel confirms message " + message.headers["seqNo"]
				);

				delete unconfirmed[event["correlationId"]];
			}
        }
        
		override public function receive( msg:IMessage, ...rest:Array) : void {
			if (msg is ReliableServerMessage) {
				if (Log.isDebug()) logger.debug(msg.body as String);
				var seqNo : Number = Number(msg.headers["seqNo"]);
				
				received.push( msg.clientId + "|"+ seqNo);
				
				if (lastServedNumber[msg.clientId]== null) {
					lastServedNumber[msg.clientId]= -1;
				}
				if ( seqNo == lastServedNumber[msg.clientId] + 1) {
						if (Log.isDebug()) logger.debug(
							"Letting out incoming message " + seqNo
						);
						super.receive( msg, rest);
						lastServedNumber[msg.clientId]++;						
						seqNo = findAdjacentBufferedMessages(msg.clientId, seqNo);
				} else if ( seqNo > lastServedNumber[msg.clientId] ) {
					if (Log.isDebug()) logger.debug(
							"Buffering message " + seqNo + " as out of order"
					);
					incoming[msg.clientId + '|' + seqNo] = {msg:msg, rest:rest};
				}
			} else
				super.receive( msg, rest);
		}
	
		private function findAdjacentBufferedMessages(clientId:String, seqNo:Number): Number {
			var more:Boolean;
			// We just processed, say, 3rd message. We may have buffered
			// 5,4,6,7.  Internal "for" loop will find 4, then external
			// "while" loop will restart the search and pick 5,6,7
			do {
				more = false;
				for each(var envelope:Object in incoming){
					var msg:IMessage = envelope.msg;					 
					if (msg.clientId != clientId)
						continue;
					if (msg.headers["seqNo"] == seqNo + 1) {
						seqNo++;
			 			lastServedNumber[clientId]++;
						if (Log.isDebug()) logger.debug(
							"Yanking message " + seqNo + " out of the buffer"
						);
						super.receive( msg, envelope.rest);
						delete incoming[seqNo];
						more = true;
					}
				}
			} while (more);
			return seqNo;
		}
	
		public static function getConfirmations(result:Array=null):Array {
			if (result==null) result = [];
			for (var i:int=0;  i<received.length; i++) {
				result.push(received[i]);
			}
			received=[];
			return result;
		}

	
		private var logger:ILogger = Log.getLogger( 
			"com.farata.messaging.channels.QoSRTMPChannel"
		);
	    public static var lastServedNumber:Dictionary  = new Dictionary();
		private static var received:Array=[];
		private  var incoming:Dictionary ;
		private  var unconfirmed:Dictionary ;

	}
}
