package com.farata.messaging.channels {
	import com.farata.messaging.messages.ReliableClientMessage;
	
	import flash.events.IOErrorEvent;
	import flash.utils.Dictionary;
	import flash.utils.setInterval;
	import flash.utils.setTimeout;
	
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.messaging.MessageAgent;
	import mx.messaging.MessageResponder;
	import mx.messaging.channels.RTMPChannel;
	import mx.messaging.events.MessageAckEvent;
	import mx.messaging.events.MessageEvent;
	import mx.messaging.messages.IMessage;

	public class ResendingRTMPChannel extends mx.messaging.channels.RTMPChannel	{
		private  var unconfirmed:Dictionary ;
		public static const RESEND_TIMEOUT:int = 3000;   // Resend unconfirmed message after 3 sec
		public static const RESENDER_SLEEP_INTERVAL:int = 500; // 

		public function ResendingRTMPChannel(id:String=null, uri:String=null) {
			super(id, uri);
			unconfirmed = new Dictionary();
			setInterval( resend, RESENDER_SLEEP_INTERVAL);
		}

		override protected function ioErrorHandler( event :IOErrorEvent ) : void {
			super.ioErrorHandler( event );
			setTimeout(resend, 1);
		}

		public function resend() : void {
			for each (var envelope:Object in unconfirmed) {
				if ( new Date().valueOf() - envelope.registeredTs > RESEND_TIMEOUT) {
					var message:IMessage = envelope.message;
					message.headers["duplicate"]="true";  // non-functional, for tracing only
					send( envelope.agent, message);
				}
			}
		}

		override public function send( agent : MessageAgent, message:IMessage) : void {
			if (message is ReliableClientMessage) {
				 unconfirmed[ message.messageId] = {
				 	message:message,
					registeredTs: new Date().valueOf(),
					agent:agent
				 };
				// Emulate 20% of "lost" messages
				if (( message.headers["tm"] != null) && (Math.random()<.2)) {
					if (Log.isDebug()) logger.debug(
						"ResendingChannel emulates loss of message " + message.headers["seqNo"]
					);
					return;
				} else { 
					if (Log.isDebug()) logger.debug(
						"ResendingChannel sends through message " + message.headers["seqNo"]
					);
					super.send( agent, message );
				}
			} else 
				super.send( agent, message );
		}

		private var _defaultAgentListener:Dictionary = new Dictionary();
		override   protected function getDefaultMessageResponder(agent:MessageAgent, msg:IMessage):MessageResponder
    	{
    		if (msg is ReliableClientMessage) {
    			if (agent != null && _defaultAgentListener[agent] == null ) {
    				_defaultAgentListener[agent] = agent;
    				agent.addEventListener(MessageAckEvent.ACKNOWLEDGE,  onProducerAcknowledge );
    			}
    		}
        	return super.getDefaultMessageResponder(agent, msg);
    	}
		
		private function onProducerAcknowledge(event:MessageEvent):void{
			var ackEvent:MessageAckEvent = event as MessageAckEvent;
 			var message:ReliableClientMessage = unconfirmed[ackEvent.correlationId].message; 
			if (Log.isDebug()) logger.debug(
				"ResendingChannel confirms message " + message.headers["seqNo"]
			);

			delete unconfirmed[ackEvent.correlationId];
        }
       	private var logger:ILogger = Log.getLogger( 
			"com.farata.messaging.channels.ResendingRTMPChannel"
		);

	}
}
