package com.farata.messaging.channels {
	import com.farata.messaging.messages.ReliableServerMessage;
	import flash.utils.Dictionary;
	import mx.logging.Log;
	import mx.logging.ILogger;
	import mx.messaging.channels.RTMPChannel;
	import mx.messaging.messages.IMessage;

  public class SerializingRTMPChannel extends mx.messaging.channels.RTMPChannel {
	public function SerializingRTMPChannel(id:String=null, uri:String=null)	{
		super(id, uri);
		incoming = new Dictionary();
	}
	
	override public function receive( msg:IMessage, ...rest:Array) : void {
		if (msg is ReliableServerMessage) {
			if (Log.isDebug()) logger.debug(msg.body as String);
			var seqNo : Number = Number(msg.headers["seqNo"]);
			
			received.push( msg.clientId + "|"+ seqNo);
			
			if (lastServedNumber[msg.clientId]== null) {
				lastServedNumber[msg.clientId]= -1;
			}
			if ( seqNo == lastServedNumber[msg.clientId] + 1) {
					if (Log.isDebug()) logger.debug(
						"Letting out incoming message " + seqNo
					);
					super.receive( msg, rest);
					lastServedNumber[msg.clientId]++;						
					seqNo = findAdjacentBufferedMessages(msg.clientId, seqNo);
			} else if ( seqNo > lastServedNumber[msg.clientId] ) {
				if (Log.isDebug()) logger.debug(
						"Buffering message " + seqNo + " as out of order"
				);
				incoming[msg.clientId + '|' + seqNo] = {msg:msg, rest:rest};
			}
		} else
			super.receive( msg, rest);
	}

	private function findAdjacentBufferedMessages(clientId:String, seqNo:Number): Number {
		var more:Boolean;
		// We just processed, say, 3rd message. We may have buffered
		// 5,4,6,7.  Internal "for" loop will find 4, then external
		// "while" loop will restart the search and pick 5,6,7
		do {
			more = false;
			for each(var envelope:Object in incoming){
				var msg:IMessage = envelope.msg;					 
				if (msg.clientId != clientId)
					continue;
				if (msg.headers["seqNo"] == seqNo + 1) {
					seqNo++;
		 			lastServedNumber[clientId]++;
					if (Log.isDebug()) logger.debug(
						"Yanking message " + seqNo + " out of the buffer"
					);
					super.receive( msg, envelope.rest);
					delete incoming[seqNo];
					more = true;
				}
			}
		} while (more);
		return seqNo;
	}

	public static function getConfirmations(result:Array=null):Array {
		if (result==null) result = [];
		for (var i:int=0;  i<received.length; i++) {
			result.push(received[i]);
		}
		received=[];
		return result;
	}


	private var logger:ILogger = Log.getLogger( 
			"com.farata.messaging.channels.SerializingRTMPChannel"
	);
    public static var lastServedNumber:Dictionary  = new Dictionary();
	private static var received:Array=[];
	private  var incoming:Dictionary ;

  }
}
