package com.farata.messaging.adapters;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.Logger;

import com.farata.messaging.messages.ClientHeartbeatMessage;
import com.farata.messaging.messages.ReliableServerMessage;

import flex.messaging.config.ConfigMap;
import flex.messaging.messages.Message;
import flex.messaging.services.MessageService;
import flex.messaging.services.messaging.adapters.MessagingAdapter;

public class QoSAdapter extends MessagingAdapter
{
	public void initialize(String id, ConfigMap properties){
		super.initialize(id, properties);

		if( resender == null) {
			resender = new Resender();
			Thread resenderThread =	new Thread(resender, "Resender");
			resenderThread.setDaemon(true);
			resenderThread.start();
		}
	}

	public Object invoke(Message message)
    {
		isDebug = logger.isDebugEnabled();
		if ( message instanceof ClientHeartbeatMessage ) {
    		processDeliveryConfirmations((ClientHeartbeatMessage) message);
    	} else if ( message instanceof ReliableServerMessage ) {
			registerForDeliveryConfirmation((ReliableServerMessage)message);

			double random = Math.random();
			if ((random<0.2) && (String)message.getHeader("tm")!=null) {
				String seqNo = (String)message.getHeader("seqNo");
				if (isDebug) logger.debug("QoS adapter emulating loss of message " + seqNo);
			} else {
				sendToClient((ReliableServerMessage)message);
			}
	   	}   	
        return null;    
    }

	private void registerForDeliveryConfirmation( ReliableServerMessage message) {
		String clientId = (String)message.getClientId();
		String seqNo = (String)message.getHeader("seqNo");
		synchronized(unconfirmedMessageMap) {
			message.setHeader("registeredTs", System.currentTimeMillis());
			unconfirmedMessageMap.put(clientId + "|" + seqNo, message);
		}
	}
		
	private void sendToClient(ReliableServerMessage message) {
		String seqNo = (String)message.getHeader("seqNo");
		String clientId = (String)message.getClientId();

		MessageService msgService = (MessageService)getDestination().getService();
		if (isDebug) logger.debug("QoS adapter is sending through message " + seqNo );
		Set<String> subscriberIds = new HashSet<String>();
		subscriberIds.add(clientId);
		msgService.pushMessageToClients(subscriberIds, message, false);
	}
		
    private void processDeliveryConfirmations(ClientHeartbeatMessage message) {
		if ((message.received!=null) && (message.received.length>0)) {
			if (isDebug) logger.debug("QoS adapter received delivery confirmations:");
			synchronized(unconfirmedMessageMap) {
				for (int i=0; i <message.received.length; i++) {
					if (isDebug) logger.debug("	...and removes (CLIENTID|seqNo)" + message.received[i]);
					unconfirmedMessageMap.remove(message.received[i]);
				}
			}
		}	
	}
    

	private boolean isDebug;
	private Resender resender = null;

	static public ConcurrentHashMap<String, ReliableServerMessage> unconfirmedMessageMap;
	static Logger logger;
	static {
		unconfirmedMessageMap = new ConcurrentHashMap<String, ReliableServerMessage>();
        logger = Logger.getLogger(QoSAdapter.class);
	}

	
}