package com.farata.messaging.adapters;

import com.farata.messaging.messages.ReliableServerMessage;
import flex.messaging.MessageBroker;
import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;
import org.apache.log4j.Logger;

public class Resender implements Runnable {
	public static int RESENDER_THREAD_SLEEP = 500;
	public static int RESEND_TIMEOUT = 3000; // Resend unconfirmed message after 3 sec
	public static int DEAD_CLIENT_TIMEOUT = 20000; // Remove permanently after 20 sec 

	protected static Resender resender = null;
	 public void run() {
		 ConcurrentHashMap<String, ReliableServerMessage> map = QoSAdapter.unconfirmedMessageMap;

		 while (true) {
	    	try {
	    		Thread.sleep(RESENDER_THREAD_SLEEP);
	    		synchronized (map) {
	        		for (Enumeration<String> e = map.keys(); e.hasMoreElements();) {
	    
	        			String key = e.nextElement();
	        			ReliableServerMessage message = map.get(key);
	
	    				String seqNo = (String)message.getHeader("seqNo");
	        			long nowTs=System.currentTimeMillis();
	        			long createdTs = message.getTimestamp();
	        			long registeredTs = (Long)message.getHeader("registeredTs");		        		
	    				
	        			if ((nowTs - createdTs) > DEAD_CLIENT_TIMEOUT) {
	        				if (logger.isDebugEnabled()) logger.debug(
	        					"Resender thread deletes message  " + seqNo 
	        				);
	        				map.remove(key);
	        			} else if ((nowTs - registeredTs) > RESEND_TIMEOUT) {
	    	        		MessageBroker mb = MessageBroker.getMessageBroker(null);
	        				if (logger.isDebugEnabled()) logger.debug(
	        					"Resender thread resends message " + seqNo 
	        				);		
	    	        		mb.routeMessageToService(message, null);
	        			}
	        		} //for
	    		} // synchronized
	    	} catch (InterruptedException ex){
	    		if (logger.isInfoEnabled()) logger.info(
	    				"...in Resender...Interrupted"
	    		);
	    	}
	    }
	}
	
	static Logger logger;
	static{
	        logger = Logger.getLogger(Resender.class);
	}
}
