package com.farata.messaging.endpoints;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.Logger;

import com.farata.messaging.messages.ReliableClientMessage;

import flex.messaging.FlexContext;
import flex.messaging.FlexSession;
import flex.messaging.endpoints.RTMPEndpoint;
import flex.messaging.messages.AcknowledgeMessage;
import flex.messaging.messages.Message;

public class SerializingRTMPEndpoint extends RTMPEndpoint {
	private final String INCOMING_MESSAGES="incomingMessages";
	private final String LAST_SERVED_NUMBER="lastServedNumber";
	private final String SEQUENCE_NUMBER="seqNo";

	@SuppressWarnings("unchecked")
	public Message serviceMessage(Message message)    {
	   	logger.info(message);

	   	if ( message instanceof ReliableClientMessage ) {
			FlexSession session =  FlexContext.getFlexSession();
			AcknowledgeMessage acknowledgeMessage = null;
	   		acknowledgeMessage = new AcknowledgeMessage();
			acknowledgeMessage.setClientId(message.getClientId());
			acknowledgeMessage.setCorrelationId(message.getMessageId());

			if (session.getAttribute(INCOMING_MESSAGES) == null)
				session.setAttribute( INCOMING_MESSAGES, new ConcurrentHashMap<Integer,Message>());
			ConcurrentHashMap<Integer,Message> incoming = (ConcurrentHashMap<Integer,Message>) session.getAttribute(INCOMING_MESSAGES);

			int lastServedNumber = -1;
			boolean isDebug=true;logger.isDebugEnabled();

			if ( session.getAttribute(LAST_SERVED_NUMBER) != null )
				lastServedNumber = (Integer)session.getAttribute(LAST_SERVED_NUMBER);
			int seqNo = (Integer)message.getHeader(SEQUENCE_NUMBER) ;
			String duplicate = (String)message.getHeader("duplicate");
			if ((duplicate!=null) && isDebug)

				if (isDebug) logger.debug(
					"Client sent duplicate to compensate send failure "+ seqNo
				);
			if (seqNo <= lastServedNumber) {
				if (isDebug) logger.debug(
					"Ignoring message " + seqNo + " as already processed"
				);
			} else if (seqNo == lastServedNumber+1){
				if (isDebug) logger.debug(
					"Letting out incoming message " + seqNo 
				);
				super.serviceMessage(message);
				lastServedNumber++;
				while ((message=(Message)incoming.remove((lastServedNumber+1)))!=null) {
					seqNo++;
					if (isDebug) logger.debug(
						"Yanking message " + seqNo + " out of the buffer" 
					);
					super.serviceMessage(message);
					lastServedNumber++;
				}
			} else  {
				if (isDebug) logger.debug(
					"Buffering message " + seqNo + " as out of order" 
				);
				incoming.put(seqNo, message);
			}
			session.setAttribute(LAST_SERVED_NUMBER, (Integer)lastServedNumber);
			acknowledgeMessage.setHeader(LAST_SERVED_NUMBER, (Integer)lastServedNumber); // Not used so far
	 		return acknowledgeMessage;
	   	} else
			return super.serviceMessage(message);
	}


	static Logger logger;
	static {
	        logger = Logger.getLogger(SerializingRTMPEndpoint.class);
	}
}
